"""Contains the base class :class:`.BaseListPrompt` which can be used to create a prompt involving choices."""
from abc import abstractmethod
from typing import Any, Callable, List, Optional

from prompt_toolkit.filters.base import Condition
from prompt_toolkit.keys import Keys

from InquirerPy.base.complex import BaseComplexPrompt
from InquirerPy.base.control import InquirerPyUIListControl
from InquirerPy.separator import Separator
from InquirerPy.utils import (
    InquirerPyKeybindings,
    InquirerPyMessage,
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
)


class BaseListPrompt(BaseComplexPrompt):
    """A base class to create a complex prompt involving choice selections (i.e. list) using `prompt_toolkit` Application.

    Note:
        This class does not create :class:`~prompt_toolkit.layout.Layout` nor :class:`~prompt_toolkit.application.Application`,
        it only contains the necessary attributes and helper functions to be consumed.

    See Also:
        :class:`~InquirerPy.prompts.list.ListPrompt`
        :class:`~InquirerPy.prompts.fuzzy.FuzzyPrompt`
    """

    def __init__(
        self,
        message: InquirerPyMessage,
        style: Optional[InquirerPyStyle] = None,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        instruction: str = "",
        long_instruction: str = "",
        border: bool = False,
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        multiselect: bool = False,
        keybindings: Optional[InquirerPyKeybindings] = None,
        cycle: bool = True,
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        super().__init__(
            message=message,
            style=style,
            border=border,
            vi_mode=vi_mode,
            qmark=qmark,
            amark=amark,
            transformer=transformer,
            filter=filter,
            invalid_message=invalid_message,
            validate=validate,
            instruction=instruction,
            long_instruction=long_instruction,
            wrap_lines=wrap_lines,
            raise_keyboard_interrupt=raise_keyboard_interrupt,
            mandatory=mandatory,
            mandatory_message=mandatory_message,
            session_result=session_result,
        )

        self._content_control: InquirerPyUIListControl
        self._multiselect = multiselect
        self._is_multiselect = Condition(lambda: self._multiselect)
        self._cycle = cycle

        if not keybindings:
            keybindings = {}

        self.kb_maps = {
            "down": [
                {"key": "down"},
                {"key": "c-n", "filter": ~self._is_vim_edit},
                {"key": "j", "filter": self._is_vim_edit},
            ],
            "up": [
                {"key": "up"},
                {"key": "c-p", "filter": ~self._is_vim_edit},
                {"key": "k", "filter": self._is_vim_edit},
            ],
            "toggle": [
                {"key": "space"},
            ],
            "toggle-down": [
                {"key": Keys.Tab},
            ],
            "toggle-up": [
                {"key": Keys.BackTab},
            ],
            "toggle-all": [
                {"key": "alt-r"},
                {"key": "c-r"},
            ],
            "toggle-all-true": [
                {"key": "alt-a"},
                {"key": "c-a"},
            ],
            "toggle-all-false": [],
            **keybindings,
        }

        self.kb_func_lookup = {
            "down": [{"func": self._handle_down}],
            "up": [{"func": self._handle_up}],
            "toggle": [{"func": self._handle_toggle_choice}],
            "toggle-down": [
                {"func": self._handle_toggle_choice},
                {"func": self._handle_down},
            ],
            "toggle-up": [
                {"func": self._handle_toggle_choice},
                {"func": self._handle_up},
            ],
            "toggle-all": [{"func": self._handle_toggle_all}],
            "toggle-all-true": [{"func": self._handle_toggle_all, "args": [True]}],
            "toggle-all-false": [{"func": self._handle_toggle_all, "args": [False]}],
        }

    @property
    def content_control(self) -> InquirerPyUIListControl:
        """Get the content controller object.

        Needs to be an instance of :class:`~InquirerPy.base.control.InquirerPyUIListControl`.

        Each :class:`.BaseComplexPrompt` requires a `content_control` to display custom
        contents for the prompt.

        Raises:
            NotImplementedError: When `self._content_control` is not found.
        """
        if not self._content_control:
            raise NotImplementedError
        return self._content_control

    @content_control.setter
    def content_control(self, value: InquirerPyUIListControl) -> None:
        self._content_control = value

    @property
    def result_name(self) -> Any:
        """Get the result value that should be printed to the terminal.

        In multiselect scenario, return result as a list.
        """
        if self._multiselect:
            return [choice["name"] for choice in self.selected_choices]
        else:
            try:
                return self.content_control.selection["name"]
            except IndexError:
                return ""

    @property
    def result_value(self) -> Any:
        """Get the result value that should return to the user.

        In multiselect scenario, return result as a list.
        """
        if self._multiselect:
            return [choice["value"] for choice in self.selected_choices]
        else:
            try:
                return self.content_control.selection["value"]
            except IndexError:
                return ""

    @property
    def selected_choices(self) -> List[Any]:
        """List[Any]: Get all user selected choices."""

        def filter_choice(choice):
            return not isinstance(choice, Separator) and choice["enabled"]

        return list(filter(filter_choice, self.content_control.choices))

    def _handle_down(self, _) -> bool:
        """Handle event when user attempts to move down.

        Returns:
            Boolean indicating if the action hits the cap.
        """
        if self._cycle:
            self.content_control.selected_choice_index = (
                self.content_control.selected_choice_index + 1
            ) % self.content_control.choice_count
            return False
        else:
            self.content_control.selected_choice_index += 1
            if (
                self.content_control.selected_choice_index
                >= self.content_control.choice_count
            ):
                self.content_control.selected_choice_index = (
                    self.content_control.choice_count - 1
                )
                return True
            return False

    def _handle_up(self, _) -> bool:
        """Handle event when user attempts to move up.

        Returns:
            Boolean indicating if the action hits the cap.
        """
        if self._cycle:
            self.content_control.selected_choice_index = (
                self.content_control.selected_choice_index - 1
            ) % self.content_control.choice_count
            return False
        else:
            self.content_control.selected_choice_index -= 1
            if self.content_control.selected_choice_index < 0:
                self.content_control.selected_choice_index = 0
                return True
            return False

    @abstractmethod
    def _handle_toggle_choice(self, event) -> None:
        """Handle event when user attempting to toggle the state of the chocie."""
        pass

    @abstractmethod
    def _handle_toggle_all(self, event, value: bool) -> None:
        """Handle event when user attempting to alter the state of all choices."""
        pass



from typing import Dict


def get_sample_style(val=None) -> Dict[str, str]:
    """For testing styles."""
    if not val:
        val = {}
    return {
        "questionmark": "#e5c07b",
        "answermark": "#e5c07b",
        "answer": "#61afef",
        "input": "#98c379",
        "question": "",
        "answered_question": "",
        "instruction": "#abb2bf",
        "long_instruction": "#abb2bf",
        "pointer": "#61afef",
        "checkbox": "#98c379",
        "separator": "",
        "skipped": "#5c6370",
        "marker": "#e5c07b",
        "validator": "",
        "fuzzy_prompt": "#c678dd",
        "fuzzy_info": "#abb2bf",
        "frame.border": "#4b5263",
        "fuzzy_match": "#c678dd",
        "spinner_pattern": "#e5c07b",
        "spinner_text": "",
        "bottom-toolbar": "noreverse",
        **val,
    }



# NOTE: Following example requires boto3 package.
import boto3

from InquirerPy import prompt
from InquirerPy.exceptions import InvalidArgument
from InquirerPy.validator import PathValidator

client = boto3.client("s3")


def get_bucket(_):
    return [bucket["Name"] for bucket in client.list_buckets()["Buckets"]]


def walk_s3_bucket(result):
    response = []
    paginator = client.get_paginator("list_objects")
    for result in paginator.paginate(Bucket=result["bucket"]):
        for file in result["Contents"]:
            response.append(file["Key"])
    return response


def is_upload(result):
    return result[0] == "Upload"


questions = [
    {
        "message": "Select an S3 action:",
        "type": "list",
        "choices": ["Upload", "Download"],
    },
    {
        "message": "Enter the filepath to upload:",
        "type": "filepath",
        "when": is_upload,
        "validate": PathValidator(),
        "only_files": True,
    },
    {
        "message": "Select a bucket:",
        "type": "fuzzy",
        "choices": get_bucket,
        "name": "bucket",
        "spinner_enable": True,
    },
    {
        "message": "Select files to download:",
        "type": "fuzzy",
        "when": lambda _: not is_upload(_),
        "choices": walk_s3_bucket,
        "multiselect": True,
        "spinner_enable": True,
    },
    {
        "message": "Enter destination folder:",
        "type": "filepath",
        "when": lambda _: not is_upload(_),
        "only_directories": True,
        "validate": PathValidator(),
    },
    {"message": "Confirm?", "type": "confirm", "default": False},
]

try:
    result = prompt(questions, vi_mode=True)
except InvalidArgument:
    print("No available choices")

# Download or Upload the file based on result ...



import os
import unittest
from functools import partial
from unittest.mock import ANY, call, patch

from prompt_toolkit.buffer import ValidationState
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.keys import Keys

from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT
from InquirerPy.exceptions import RequiredKeyNotFound
from InquirerPy.prompts.input import InputPrompt
from InquirerPy.utils import get_style
from InquirerPy.validator import NumberValidator
from tests.style import get_sample_style


class TestBaseSimple(unittest.TestCase):
    @patch("InquirerPy.base.simple.KeyBindings.add")
    @patch("InquirerPy.base.simple.Validator.from_callable")
    @patch("InquirerPy.base.simple.Style.from_dict")
    def test_constructor_default(self, mocked_style, mocked_validator, mocked_kb):
        input_prompt = InputPrompt(message="Enter your name", style=None, default="1")
        self.assertEqual(input_prompt._message, "Enter your name")
        mocked_style.assert_has_calls([call(get_sample_style())])
        self.assertEqual(input_prompt._default, "1")
        self.assertEqual(input_prompt._qmark, "?")
        self.assertEqual(input_prompt._amark, "?")
        self.assertEqual(input_prompt._editing_mode, EditingMode.EMACS)
        mocked_validator.assert_has_calls(
            [call(ANY, "Invalid input", move_cursor_to_end=True)]
        )
        mocked_kb.assert_has_calls([call("c-c")])

    @patch("InquirerPy.base.simple.Validator.from_callable")
    @patch("InquirerPy.base.simple.Style.from_dict")
    def test_constructor_custom(self, mocked_style, mocked_validator):
        input_prompt = InputPrompt(
            message=lambda _: "Enter your name",
            style=get_style({"questionmark": "#111111"}, style_override=False),
            qmark="[?]",
            amark="*",
            default=lambda _: "1",
            vi_mode=True,
            validate=NumberValidator(),
        )
        style = get_sample_style()
        style["questionmark"] = "#111111"
        self.assertEqual(input_prompt._message, "Enter your name")
        mocked_style.assert_has_calls([call(style)])
        self.assertEqual(input_prompt._default, "1")
        self.assertEqual(input_prompt._qmark, "[?]")
        self.assertEqual(input_prompt._amark, "*")
        self.assertEqual(input_prompt._editing_mode, EditingMode.VI)
        mocked_validator.assert_not_called()

    def test_vi_kb(self):
        prompt = InputPrompt(message="")
        self.assertEqual(prompt._editing_mode, EditingMode.EMACS)
        os.environ["INQUIRERPY_VI_MODE"] = "true"
        prompt = InputPrompt(message="")
        self.assertEqual(prompt._editing_mode, EditingMode.VI)
        prompt = InputPrompt(message="", vi_mode=False)
        self.assertEqual(prompt._editing_mode, EditingMode.VI)
        del os.environ["INQUIRERPY_VI_MODE"]

    def test_prompt_message_initial(self):
        input_prompt = InputPrompt(message="Enter your name", style=None, qmark="[?]")
        message = input_prompt._get_prompt_message()
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " Enter your name"),
                ("class:instruction", " "),
            ],
        )

    def test_prompt_message_answered(self):
        input_prompt = InputPrompt(message="Enter your name", style=None, qmark="[?]")
        input_prompt.status["answered"] = True
        input_prompt.status["result"] = "haha"
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Enter your name"),
                ("class:answer", " haha"),
            ],
        )

    def test_prompt_message_kbi(self):
        input_prompt = InputPrompt(message="Enter your name", style=None, qmark="[?]")
        input_prompt.status["answered"] = True
        input_prompt.status["result"] = INQUIRERPY_KEYBOARD_INTERRUPT
        input_prompt.status["skipped"] = True
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message, [("class:skipped", "[?]"), ("class:skipped", " Enter your name ")]
        )

    @patch("InquirerPy.base.simple.KeyBindings.add")
    def test_register_kb(self, mocked_kb):
        prompt = InputPrompt(message="")
        hello = prompt.register_kb("alt-a", "alt-b", filter=True)

        @hello
        def _(_):  # type:ignore
            pass

        mocked_kb.assert_has_calls([call("escape", "a", "escape", "b", filter=True)])

        condition = Condition(lambda: True)
        hello = prompt.register_kb("c-i", filter=condition)

        @hello
        def _(_):
            pass

        mocked_kb.assert_has_calls([call("c-i", filter=condition)])

    @patch("InquirerPy.base.simple.BaseSimplePrompt.register_kb")
    def test_keybinding_factory(self, mocked_kb):
        prompt = InputPrompt(message="")
        mocked_kb.assert_has_calls([call(Keys.Enter, filter=ANY)])
        mocked_kb.assert_has_calls([call(Keys.Escape, Keys.Enter, filter=ANY)])
        mocked_kb.assert_has_calls([call("c-c", filter=prompt._is_rasing_kbi)])
        mocked_kb.assert_has_calls([call("c-d", filter=~prompt._is_rasing_kbi)])
        mocked_kb.assert_has_calls([call("c-c", filter=~prompt._is_rasing_kbi)])
        mocked_kb.reset_mock()
        prompt = partial(
            InputPrompt, message="", keybindings={"hello": [{"key": "c-d"}]}
        )
        self.assertRaises(RequiredKeyNotFound, prompt)

    def test_handle_interrupt(self):
        prompt = InputPrompt(message="")
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_interrupt(event)
        self.assertEqual(prompt.status["answered"], True)
        self.assertEqual(prompt.status["result"], INQUIRERPY_KEYBOARD_INTERRUPT)
        self.assertEqual(prompt.status["skipped"], True)

    @patch.object(InputPrompt, "_run")
    def test_execute_kbi(self, mocked_run):
        prompt = InputPrompt(message="")
        mocked_run.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        self.assertTrue(prompt._raise_kbi)
        self.assertRaises(KeyboardInterrupt, prompt.execute, True)

        os.environ["INQUIRERPY_NO_RAISE_KBI"] = "True"
        prompt = InputPrompt(message="")
        self.assertFalse(prompt._raise_kbi)
        del os.environ["INQUIRERPY_NO_RAISE_KBI"]

    @patch.object(InputPrompt, "_run")
    def test_execute_filter(self, mocked_run):
        mocked_run.return_value = "1"
        prompt = InputPrompt(message="")
        result = prompt.execute()
        self.assertEqual(result, "1")

        prompt = InputPrompt(message="", filter=lambda result: int(result) * 2)
        result = prompt.execute()
        self.assertEqual(result, 2)

    def test_handle_skip(self) -> None:
        prompt = InputPrompt(message="", mandatory=True, mandatory_message="hello")
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_skip(event)

        self.assertEqual(
            prompt._session.default_buffer.validation_state, ValidationState.INVALID
        )
        self.assertEqual(str(prompt._session.default_buffer.validation_error), "hello")
        self.assertEqual(prompt.status["answered"], False)
        self.assertEqual(prompt.status["skipped"], False)
        self.assertEqual(prompt.status["result"], None)

        prompt = InputPrompt(message="", mandatory=False)
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_skip(event)
        self.assertEqual(prompt.status["answered"], True)
        self.assertEqual(prompt.status["skipped"], True)
        self.assertEqual(prompt.status["result"], None)



import unittest
from unittest.mock import ANY, call, patch

from InquirerPy.base import BaseComplexPrompt
from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.prompts.rawlist import InquirerPyRawlistControl, RawlistPrompt
from InquirerPy.separator import Separator


class TestRawList(unittest.TestCase):
    choices = [
        {"name": "foo", "value": "boo", "enabled": True},
        "hello",
        Separator(),
        "yes",
    ]

    def test_content_control(self):
        content_control = InquirerPyRawlistControl(
            self.choices, "yes", " ", ") ", ">", None, True, " "
        )
        self.assertEqual(content_control._pointer, " ")
        self.assertEqual(content_control._separator, ") ")
        self.assertEqual(content_control.choice_count, 4)
        self.assertEqual(content_control.selected_choice_index, 3)
        self.assertEqual(
            content_control._get_hover_text(content_control.choices[0]),
            [
                ("class:pointer", " "),
                ("class:marker", ">"),
                ("class:pointer", "1) "),
                ("[SetCursorPosition]", ""),
                ("class:pointer", "foo"),
            ],
        )
        self.assertEqual(
            content_control._get_normal_text(content_control.choices[0]),
            [("", " "), ("class:marker", ">"), ("", "1) "), ("", "foo")],
        )
        self.assertEqual(
            content_control.choices,
            [
                {
                    "actual_index": 0,
                    "display_index": 1,
                    "name": "foo",
                    "value": "boo",
                    "enabled": True,
                },
                {
                    "actual_index": 1,
                    "display_index": 2,
                    "name": "hello",
                    "value": "hello",
                    "enabled": False,
                },
                {"name": "---------------", "value": ANY, "enabled": False},
                {
                    "actual_index": 3,
                    "display_index": 3,
                    "name": "yes",
                    "value": "yes",
                    "enabled": False,
                },
            ],
        )
        self.assertEqual(
            content_control._get_formatted_choices(),
            [
                ("", " "),
                ("class:marker", ">"),
                ("", "1) "),
                ("", "foo"),
                ("", "\n"),
                ("", " "),
                ("class:marker", " "),
                ("", "2) "),
                ("", "hello"),
                ("", "\n"),
                ("", " "),
                ("class:marker", " "),
                ("class:separator", "---------------"),
                ("", "\n"),
                ("class:pointer", " "),
                ("class:marker", " "),
                ("class:pointer", "3) "),
                ("[SetCursorPosition]", ""),
                ("class:pointer", "yes"),
            ],
        )

        content_control = InquirerPyRawlistControl(
            self.choices, 2, " ", ")", ">", None, False, " "
        )
        self.assertEqual(content_control.selected_choice_index, 1)

    def test_content_control_exceptions(self):
        choices = [{"hello": "hello"}]
        self.assertRaises(
            RequiredKeyNotFound,
            InquirerPyRawlistControl,
            choices,
            "",
            "",
            "",
            "",
            None,
            False,
            " ",
        )

        choices = [Separator(), Separator()]
        self.assertRaises(
            InvalidArgument,
            InquirerPyRawlistControl,
            choices,
            "",
            "",
            "",
            "",
            None,
            True,
            " ",
        )

        choices = []
        self.assertRaises(
            InvalidArgument,
            InquirerPyRawlistControl,
            choices,
            "",
            "",
            "",
            "",
            None,
            False,
            " ",
        )

    def test_prompt(self):
        rawlist_prompt = RawlistPrompt(
            message="hello",
            choices=self.choices,
            default="hello",
            separator=".",
            instruction="bb",
        )
        self.assertEqual(rawlist_prompt.instruction, "bb")
        self.assertEqual(rawlist_prompt._message, "hello")

    def test_minimum_args(self):
        RawlistPrompt(message="what", choices=self.choices)

    def test_prompt_message(self):
        prompt = RawlistPrompt(
            message="hello",
            choices=self.choices,
            default="hello",
            separator=".",
            instruction="bb",
        )
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:questionmark", "?"),
                ("class:question", " hello"),
                ("class:instruction", " bb "),
                ("class:input", "2"),
            ],
        )
        prompt.status["answered"] = True
        prompt.status["result"] = []
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " hello"),
                ("class:answer", " []"),
            ],
        )

    def test_prompt_bindings(self):
        prompt = RawlistPrompt(
            message="hello",
            choices=self.choices,
            default="hello",
            separator=".",
            instruction="bb",
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)

        self.assertEqual(prompt.status["result"], None)
        self.assertEqual(prompt.status["answered"], False)
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(prompt.status["result"], "hello")
        self.assertEqual(prompt.status["answered"], True)

    @patch.object(BaseComplexPrompt, "register_kb")
    def test_kb_added(self, mocked_add):
        prompt = RawlistPrompt(
            message="hello",
            choices=self.choices,
            default="hello",
            separator=".",
            instruction="bb",
        )
        prompt._on_rendered("")
        mocked_add.assert_has_calls([call("1")])
        mocked_add.assert_has_calls([call("2")])
        mocked_add.assert_has_calls([call("3")])

    def test_rawlist_10(self):
        prompt = RawlistPrompt(message="", choices=[i for i in range(10)])
        self.assertRaises(InvalidArgument, prompt._on_rendered, "")
        prompt = RawlistPrompt(message="", choices=[i for i in range(9)])
        prompt._after_render(None)



"""Module contains pre-built validators."""
import re
from pathlib import Path
from typing import Optional

from prompt_toolkit.validation import ValidationError, Validator

__all__ = [
    "PathValidator",
    "EmptyInputValidator",
    "PasswordValidator",
    "NumberValidator",
]


class NumberValidator(Validator):
    """:class:`~prompt_toolkit.validation.Validator` to validate if input is a number.

    Args:
        message: Error message to display in the validatation toolbar when validation failed.
        float_allowed: Allow input to contain floating number (with decimal).
    """

    def __init__(
        self, message: str = "Input should be a number", float_allowed: bool = False
    ) -> None:
        self._message = message
        self._float_allowed = float_allowed

    def validate(self, document) -> None:
        """Check if user input is a valid number.

        This method is used internally by `prompt_toolkit <https://python-prompt-toolkit.readthedocs.io/en/master/>`_.

        See Also:
            https://python-prompt-toolkit.readthedocs.io/en/master/pages/asking_for_input.html?highlight=validator#input-validation
        """
        try:
            if self._float_allowed:
                float(document.text)
            else:
                int(document.text)
        except ValueError:
            raise ValidationError(
                message=self._message, cursor_position=document.cursor_position
            )


class PathValidator(Validator):
    """:class:`~prompt_toolkit.validation.Validator` to validate if input is a valid filepath on the system.

    Args:
        message: Error message to display in the validatation toolbar when validation failed.
        is_file: Explicitly check if the input is a valid file on the system.
        is_dir: Explicitly check if the input is a valid directory/folder on the system.
    """

    def __init__(
        self,
        message: str = "Input is not a valid path",
        is_file: bool = False,
        is_dir: bool = False,
    ) -> None:
        self._message = message
        self._is_file = is_file
        self._is_dir = is_dir

    def validate(self, document) -> None:
        """Check if user input is a filepath that exists on the system based on conditions.

        This method is used internally by `prompt_toolkit <https://python-prompt-toolkit.readthedocs.io/en/master/>`_.

        See Also:
            https://python-prompt-toolkit.readthedocs.io/en/master/pages/asking_for_input.html?highlight=validator#input-validation
        """
        path = Path(document.text).expanduser()
        if self._is_file and not path.is_file():
            raise ValidationError(
                message=self._message,
                cursor_position=document.cursor_position,
            )
        elif self._is_dir and not path.is_dir():
            raise ValidationError(
                message=self._message,
                cursor_position=document.cursor_position,
            )
        elif not path.exists():
            raise ValidationError(
                message=self._message,
                cursor_position=document.cursor_position,
            )


class EmptyInputValidator(Validator):
    """:class:`~prompt_toolkit.validation.Validator` to validate if the input is empty.

    Args:
        message: Error message to display in the validatation toolbar when validation failed.
    """

    def __init__(self, message: str = "Input cannot be empty") -> None:
        self._message = message

    def validate(self, document) -> None:
        """Check if user input is empty.

        This method is used internally by `prompt_toolkit <https://python-prompt-toolkit.readthedocs.io/en/master/>`_.

        See Also:
            https://python-prompt-toolkit.readthedocs.io/en/master/pages/asking_for_input.html?highlight=validator#input-validation
        """
        if not len(document.text) > 0:
            raise ValidationError(
                message=self._message,
                cursor_position=document.cursor_position,
            )


class PasswordValidator(Validator):
    """:class:`~prompt_toolkit.validation.Validator` to validate password compliance.

    Args:
        message: Error message to display in the validatation toolbar when validation failed.
        length: The minimum length of the password.
        cap: Password should include at least one capital letter.
        special: Password should include at least one special char "@$!%*#?&".
        number: Password should include at least one number.
    """

    def __init__(
        self,
        message: str = "Input is not compliant with the password constraints",
        length: Optional[int] = None,
        cap: bool = False,
        special: bool = False,
        number: bool = False,
    ) -> None:
        password_pattern = r"^"
        if cap:
            password_pattern += r"(?=.*[A-Z])"
        if special:
            password_pattern += r"(?=.*[@$!%*#?&])"
        if number:
            password_pattern += r"(?=.*[0-9])"
        password_pattern += r"."
        if length:
            password_pattern += r"{%s,}" % length
        else:
            password_pattern += r"*"
        password_pattern += r"$"
        self._re = re.compile(password_pattern)
        self._message = message

    def validate(self, document) -> None:
        """Check if user input is compliant with the specified password constraints.

        This method is used internally by `prompt_toolkit <https://python-prompt-toolkit.readthedocs.io/en/master/>`_.

        See Also:
            https://python-prompt-toolkit.readthedocs.io/en/master/pages/asking_for_input.html?highlight=validator#input-validation
        """
        if not self._re.match(document.text):
            raise ValidationError(
                message=self._message, cursor_position=document.cursor_position
            )



import re
import unittest

from InquirerPy import inquirer
from InquirerPy.resolver import question_mapping


class TestInquirer(unittest.TestCase):
    def test_pkgs(self) -> None:
        inquirer_lookup = set()
        special_mapping = {"password": "secret", "input": "text", "list": "select"}
        for pkg in dir(inquirer):
            dunder_pattern = re.compile(r"^__.*")
            if not dunder_pattern.match(pkg):
                inquirer_lookup.add(pkg)

        for prompt in question_mapping.keys():
            prompt = special_mapping.get(prompt, prompt)
            if prompt not in inquirer_lookup:
                self.fail()



"""Module contains the class to create a list prompt."""
import shutil
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union

from prompt_toolkit.application.application import Application
from prompt_toolkit.filters.cli import IsDone
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    Window,
)
from prompt_toolkit.layout.controls import DummyControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.validation import ValidationError
from prompt_toolkit.widgets.base import Frame

from InquirerPy.base import InquirerPyUIListControl
from InquirerPy.base.complex import FakeDocument
from InquirerPy.base.list import BaseListPrompt
from InquirerPy.containers.instruction import InstructionWindow
from InquirerPy.containers.message import MessageWindow
from InquirerPy.containers.validation import ValidationFloat
from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.separator import Separator
from InquirerPy.utils import (
    InquirerPyDefault,
    InquirerPyKeybindings,
    InquirerPyListChoices,
    InquirerPyMessage,
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
    calculate_height,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

__all__ = ["ListPrompt"]


class InquirerPyListControl(InquirerPyUIListControl):
    """An :class:`~prompt_toolkit.layout.UIControl` class that displays a list of choices.

    Reference the parameter definition in :class:`.ListPrompt`.
    """

    def __init__(
        self,
        choices: InquirerPyListChoices,
        default: Any,
        pointer: str,
        marker: str,
        session_result: Optional[InquirerPySessionResult],
        multiselect: bool,
        marker_pl: str,
    ) -> None:
        self._pointer: str = pointer
        self._marker: str = marker
        self._marker_pl: str = marker_pl
        super().__init__(
            choices=choices,
            default=default,
            session_result=session_result,
            multiselect=multiselect,
        )

    def _get_hover_text(self, choice) -> List[Tuple[str, str]]:
        display_choices = []
        display_choices.append(("class:pointer", self._pointer))
        display_choices.append(
            (
                "class:marker",
                self._marker if choice["enabled"] else self._marker_pl,
            )
        )
        display_choices.append(("[SetCursorPosition]", ""))
        display_choices.append(("class:pointer", choice["name"]))
        return display_choices

    def _get_normal_text(self, choice) -> List[Tuple[str, str]]:
        display_choices = []
        display_choices.append(("", len(self._pointer) * " "))
        display_choices.append(
            (
                "class:marker",
                self._marker if choice["enabled"] else self._marker_pl,
            )
        )
        if not isinstance(choice["value"], Separator):
            display_choices.append(("", choice["name"]))
        else:
            display_choices.append(("class:separator", choice["name"]))
        return display_choices


class ListPrompt(BaseListPrompt):
    """Create a prompt that displays a list of choices to select.

    A wrapper class around :class:`~prompt_toolkit.application.Application`.

    Args:
        message: The question to ask the user.
            Refer to :ref:`pages/dynamic:message` documentation for more details.
        choices: List of choices to display and select.
            Refer to :ref:`pages/dynamic:choices` documentation for more details.
        style: An :class:`InquirerPyStyle` instance.
            Refer to :ref:`Style <pages/style:Alternate Syntax>` documentation for more details.
        vi_mode: Use vim keybinding for the prompt.
            Refer to :ref:`pages/kb:Keybindings` documentation for more details.
        default: Set the default value of the prompt.
            This will be used to determine which choice is highlighted (current selection),
            The default value should be the value of one of the choices.
            Refer to :ref:`pages/dynamic:default` documentation for more details.
        qmark: Question mark symbol. Custom symbol that will be displayed infront of the question before its answered.
        amark: Answer mark symbol. Custom symbol that will be displayed infront of the question after its answered.
        pointer: Pointer symbol. Customer symbol that will be used to indicate the current choice selection.
        instruction: Short instruction to display next to the question.
        long_instruction: Long instructions to display at the bottom of the prompt.
        validate: Add validation to user input.
            The main use case for this prompt would be when `multiselect` is True, you can enforce a min/max selection.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        invalid_message: Error message to display when user input is invalid.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        transformer: A function which performs additional transformation on the value that gets printed to the terminal.
            Different than `filter` parameter, this is only visual effect and won’t affect the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:transformer` documentation for more details.
        filter: A function which performs additional transformation on the result.
            This affects the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:filter` documentation for more details.
        height: Preferred height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        max_height: Max height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        multiselect: Enable multi-selection on choices.
            You can use `validate` parameter to control min/max selections.
            Setting to True will also change the result from a single value to a list of values.
        marker: Marker Symbol. Custom symbol to indicate if a choice is selected.
            This will take effects when `multiselect` is True.
        marker_pl: Marker place holder when the choice is not selected.
            This is empty space by default.
        border: Create border around the choice window.
        keybindings: Customise the builtin keybindings.
            Refer to :ref:`pages/kb:Keybindings` for more details.
        show_cursor: Display cursor at the end of the prompt.
            Set to False to hide the cursor.
        cycle: Return to top item if hit bottom during navigation or vice versa.
        wrap_lines: Soft wrap question lines when question exceeds the terminal width.
        raise_keyboard_interrupt: Raise the :class:`KeyboardInterrupt` exception when `ctrl-c` is pressed. If false, the result
            will be `None` and the question is skiped.
        mandatory: Indicate if the prompt is mandatory. If True, then the question cannot be skipped.
        mandatory_message: Error message to show when user attempts to skip mandatory prompt.
        session_result: Used internally for :ref:`index:Classic Syntax (PyInquirer)`.

    Examples:
        >>> from InquirerPy import inquirer
        >>> result = inquirer.select(message="Select one:", choices=[1, 2, 3]).execute()
        >>> print(result)
        1
    """

    def __init__(
        self,
        message: InquirerPyMessage,
        choices: InquirerPyListChoices,
        default: InquirerPyDefault = None,
        style: Optional[InquirerPyStyle] = None,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        pointer: str = INQUIRERPY_POINTER_SEQUENCE,
        instruction: str = "",
        long_instruction: str = "",
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        height: Optional[Union[int, str]] = None,
        max_height: Optional[Union[int, str]] = None,
        multiselect: bool = False,
        marker: str = INQUIRERPY_POINTER_SEQUENCE,
        marker_pl: str = " ",
        border: bool = False,
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        keybindings: Optional[InquirerPyKeybindings] = None,
        show_cursor: bool = True,
        cycle: bool = True,
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        if not hasattr(self, "_content_control"):
            self.content_control = InquirerPyListControl(
                choices=choices,
                default=default,
                pointer=pointer,
                marker=marker,
                session_result=session_result,
                multiselect=multiselect,
                marker_pl=marker_pl,
            )
        super().__init__(
            message=message,
            style=style,
            border=border,
            vi_mode=vi_mode,
            qmark=qmark,
            amark=amark,
            instruction=instruction,
            long_instruction=long_instruction,
            transformer=transformer,
            filter=filter,
            validate=validate,
            invalid_message=invalid_message,
            multiselect=multiselect,
            keybindings=keybindings,
            cycle=cycle,
            wrap_lines=wrap_lines,
            raise_keyboard_interrupt=raise_keyboard_interrupt,
            mandatory=mandatory,
            mandatory_message=mandatory_message,
            session_result=session_result,
        )
        self._show_cursor = show_cursor
        self._dimmension_height, self._dimmension_max_height = calculate_height(
            height, max_height, height_offset=self.height_offset
        )
        main_content_window = Window(
            content=self.content_control,
            height=Dimension(
                max=self._dimmension_max_height,
                preferred=self._dimmension_height,
            ),
            dont_extend_height=True,
        )

        if self._border:
            main_content_window = Frame(main_content_window)

        self._layout = FloatContainer(
            content=HSplit(
                [
                    MessageWindow(
                        message=self._get_prompt_message_with_cursor
                        if self._show_cursor
                        else self._get_prompt_message,
                        filter=True,
                        wrap_lines=self._wrap_lines,
                        show_cursor=self._show_cursor,
                    ),
                    ConditionalContainer(main_content_window, filter=~IsDone()),
                    ConditionalContainer(
                        Window(content=DummyControl()),
                        filter=~IsDone() & self._is_displaying_long_instruction,
                    ),
                    InstructionWindow(
                        message=self._long_instruction,
                        filter=~IsDone() & self._is_displaying_long_instruction,
                        wrap_lines=self._wrap_lines,
                    ),
                ]
            ),
            floats=[
                ValidationFloat(
                    invalid_message=self._get_error_message,
                    filter=self._is_invalid & ~IsDone(),
                    wrap_lines=self._wrap_lines,
                    left=0,
                    bottom=self._validation_window_bottom_offset,
                ),
            ],
        )

        self.application = Application(
            layout=Layout(self._layout),
            style=self._style,
            key_bindings=self._kb,
            after_render=self._after_render,
        )

    def _get_prompt_message_with_cursor(self) -> List[Tuple[str, str]]:
        """Obtain the prompt message to display and display cursor behind the message.

        This ensures that cursor is always at the end of the window.
        """
        message = self._get_prompt_message()
        message.append(("[SetCursorPosition]", ""))
        message.append(("", " "))  # [SetCursorPosition] require char behind it
        return message

    def _handle_toggle_choice(self, _) -> None:
        """Toggle the `enabled` status of the choice."""
        if not self._multiselect:
            return
        self.content_control.selection["enabled"] = not self.content_control.selection[
            "enabled"
        ]

    def _handle_toggle_all(self, _, value: Optional[bool] = None) -> None:
        """Toggle all choice `enabled` status.

        Args:
            value: Sepcify a value to toggle.
        """
        if not self._multiselect:
            return
        for choice in self.content_control.choices:
            if isinstance(choice["value"], Separator):
                continue
            choice["enabled"] = value if value else not choice["enabled"]

    def _handle_up(self, event) -> None:
        """Handle the event when user attempt to move up."""
        while True:
            cap = super()._handle_up(event)
            if not isinstance(self.content_control.selection["value"], Separator):
                break
            else:
                if cap and not self._cycle:
                    self._handle_down(event)
                    break

    def _handle_down(self, event) -> None:
        """Handle the event when user attempt to move down."""
        while True:
            cap = super()._handle_down(event)
            if not isinstance(self.content_control.selection["value"], Separator):
                break
            else:
                if cap and not self._cycle:
                    self._handle_up(event)
                    break

    def _handle_enter(self, event: "KeyPressEvent") -> None:
        """Handle the event when user hit `enter` key.

        1. Set the prompt state to answered.
        2. Set the result to the name of the selected choices.
        3. Exit the app with the value of the selected choices.

        In multiselect scenario, if nothing is selected, return the current highlighted choice.
        """
        try:
            fake_document = FakeDocument(self.result_value)
            self._validator.validate(fake_document)  # type: ignore
        except ValidationError as e:
            self._set_error(str(e))
        else:
            self.status["answered"] = True
            if self._multiselect and not self.selected_choices:
                self.status["result"] = [self.content_control.selection["name"]]
                event.app.exit(result=[self.content_control.selection["value"]])
            else:
                self.status["result"] = self.result_name
                event.app.exit(result=self.result_value)

    @property
    def extra_message_line_count(self) -> int:
        """int: Get extra lines created for message caused by line wrapping.

        Overriding it to count the cursor as well.
        """
        cursor_offset = -1 if not self._show_cursor else 0
        term_width, _ = shutil.get_terminal_size()
        return (self.total_message_length + cursor_offset) // term_width



import asyncio
import unittest
from unittest.mock import patch

from prompt_toolkit.input.defaults import create_pipe_input
from prompt_toolkit.output import DummyOutput
from prompt_toolkit.shortcuts.prompt import PromptSession

from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT
from InquirerPy.prompts.confirm import ConfirmPrompt
from InquirerPy.utils import InquirerPyStyle


class TestConfirmPrompt(unittest.TestCase):
    def setUp(self):
        self.inp = create_pipe_input()

    def tearDown(self):
        self.inp.close()

    def test_default_false(self):
        self.inp.send_text("\n")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=InquirerPyStyle({"qmark": "bold"}),
            default=False,
            qmark="x",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, False)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], False)

    def test_default_true(self):
        self.inp.send_text("\n")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=InquirerPyStyle({"qmark": "bold", "answer": "#000000"}),
            default=True,
            qmark="x",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, True)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], True)

    def test_input_y(self):
        self.inp.send_text("y")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt._session.prompt()
        self.assertEqual(result, True)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], True)

        self.inp.send_text("Y")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, True)

    def test_input_n(self):
        self.inp.send_text("n")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, False)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], False)

        self.inp.send_text("N")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, False)

    def test_custom_confirm(self):
        self.inp.send_text("s")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
            confirm_letter="s",
        )
        result = confirm_prompt._session.prompt()
        self.assertEqual(result, True)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], True)

        self.inp.send_text("S")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
            confirm_letter="s",
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, True)

    def test_custom_reject(self):
        self.inp.send_text("w")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=False,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
            reject_letter="w",
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, False)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], False)

        self.inp.send_text("W")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
            reject_letter="w",
        )
        result = confirm_prompt.execute()
        self.assertEqual(result, False)

    def test_get_prompt_message(self):
        confirm_prompt = ConfirmPrompt(
            message="hello", style=None, default=True, qmark="?", confirm_letter="W"
        )
        message = confirm_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "?"),
                ("class:question", " hello"),
                ("class:instruction", " (W/n) "),
            ],
        )

        confirm_prompt.status["answered"] = True
        confirm_prompt.status["result"] = True
        message = confirm_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "?"),
                ("class:answered_question", " hello"),
                ("class:answer", " Yes"),
            ],
        )

        confirm_prompt = ConfirmPrompt(
            message="Are you sure?",
            style=None,
            default=False,
            qmark="x",
            amark="x",
        )
        message = confirm_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "x"),
                ("class:question", " Are you sure?"),
                ("class:instruction", " (y/N) "),
            ],
        )

        confirm_prompt.status["answered"] = True
        confirm_prompt.status["result"] = False
        message = confirm_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "x"),
                ("class:answered_question", " Are you sure?"),
                ("class:answer", " No"),
            ],
        )

        # instruction
        confirm_prompt = ConfirmPrompt(
            message="Are you sure?",
            style=None,
            default=False,
            qmark="x",
            amark="x",
            instruction="(abc)",
        )
        message = confirm_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "x"),
                ("class:question", " Are you sure?"),
                ("class:instruction", " (abc) "),
            ],
        )

    @patch("InquirerPy.prompts.confirm.ConfirmPrompt._get_prompt_message")
    @patch("InquirerPy.base.simple.Style.from_dict")
    @patch("InquirerPy.base.simple.KeyBindings")
    @patch("InquirerPy.prompts.confirm.PromptSession")
    def test_callable_called(
        self, MockedSession, MockedKeyBindings, MockedStyle, mocked_message
    ):
        ConfirmPrompt(
            message="Are you sure?",
            style=None,
            default=False,
            qmark="x",
        )
        kb = MockedKeyBindings()
        style = MockedStyle()
        MockedSession.assert_called_once_with(
            message=mocked_message,
            key_bindings=kb,
            style=style,
            wrap_lines=True,
            bottom_toolbar=None,
            input=None,
            output=None,
        )

    @patch.object(PromptSession, "prompt")
    def test_confirm_kbi(self, mocked_session):
        mocked_session.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        prompt = ConfirmPrompt(message="")
        self.assertRaises(KeyboardInterrupt, prompt.execute)
        mocked_session.return_value = False
        prompt.execute()

    def test_input_y_async(self):
        self.inp.send_text("y")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = asyncio.run(confirm_prompt._run_async())
        self.assertEqual(result, True)
        self.assertEqual(confirm_prompt.status["answered"], True)
        self.assertEqual(confirm_prompt.status["result"], True)

        self.inp.send_text("Y")
        confirm_prompt = ConfirmPrompt(
            message="hello",
            style=None,
            default=True,
            qmark="?",
            output=DummyOutput(),
            input=self.inp,
        )
        result = asyncio.run(confirm_prompt.execute_async())
        self.assertEqual(result, True)



from InquirerPy import inquirer


def main():
    proceed, service, confirm = False, False, False
    proceed = inquirer.confirm(message="Proceed?", default=True).execute()
    if proceed:
        service = inquirer.confirm(message="Require 1 on 1?").execute()
    if service:
        confirm = inquirer.confirm(message="Confirm?").execute()


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import ANY, call, patch

from prompt_toolkit.validation import ValidationError, Validator

from InquirerPy.base.complex import BaseComplexPrompt
from InquirerPy.prompts.fuzzy import FuzzyPrompt


class TestBaseComplex(unittest.TestCase):
    def test_register_kb(self) -> None:
        fuzzy_prompt = FuzzyPrompt(message="hello", choices=["1", "2", "3"])

        @fuzzy_prompt.register_kb("c-q")
        def hello(_):
            pass

        self.assertFalse(fuzzy_prompt._invalid)
        fuzzy_prompt._invalid = True
        hello("")  # type: ignore
        self.assertFalse(fuzzy_prompt._invalid)

    @patch.object(BaseComplexPrompt, "register_kb")
    def test_after_render(self, mocked_kb):
        prompt = FuzzyPrompt(message="", choices=lambda _: [1, 2, 3])
        self.assertEqual(prompt._rendered, False)
        prompt._after_render(None)

        self.assertEqual(prompt._rendered, True)
        mocked_kb.assert_has_calls(
            [
                call("down", filter=ANY),
            ]
        )
        mocked_kb.assert_has_calls(
            [
                call("c-n", filter=ANY),
            ]
        )

    def test_prompt_message(self):
        prompt = FuzzyPrompt(message="Select one of them", choices=lambda _: [1, 2, 3])

        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:questionmark", "?"),
                ("class:question", " Select one of them"),
                ("class:instruction", " "),
            ],
        )
        prompt.status["answered"] = True
        prompt.status["result"] = ["hello"]
        prompt.status["skipped"] = False

        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Select one of them"),
                ("class:answer", " ['hello']"),
            ],
        )

        prompt = FuzzyPrompt(
            message=lambda result: "no" if not result else "hello",
            choices=[1, 2, 3],
            session_result={"1": True},
        )
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:questionmark", "?"),
                ("class:question", " hello"),
                ("class:instruction", " "),
            ],
        )

    def test_content_control(self):
        prompt = FuzzyPrompt(message="Select one of them", choices=lambda _: [1, 2, 3])
        prompt.content_control = None  # type: ignore
        try:
            prompt.content_control
        except NotImplementedError:
            pass
        else:
            self.fail("NotImplemented not raised.")

    def test_result_name_simple(self):
        prompt = FuzzyPrompt(message="Select one of them", choices=[1, 2, 3])
        self.assertEqual(prompt.result_name, "1")
        prompt.content_control.selected_choice_index = 1
        self.assertEqual(prompt.result_name, "2")

        prompt.content_control.selected_choice_index = 4
        self.assertEqual(prompt.result_name, "")

    def test_result_name_multi(self):
        prompt = FuzzyPrompt(
            message="Select one of them", choices=[1, 2, 3], multiselect=True
        )
        self.assertEqual(prompt.result_name, [])
        prompt.content_control.choices[0]["enabled"] = True
        prompt.content_control.choices[2]["enabled"] = True
        self.assertEqual(prompt.result_name, ["1", "3"])

    def test_result_value_simple(self):
        prompt = FuzzyPrompt(message="Select one of them", choices=[1, 2, 3])

        self.assertEqual(prompt.result_value, 1)
        prompt.content_control.selected_choice_index = 1
        self.assertEqual(prompt.result_value, 2)

        prompt.content_control.selected_choice_index = 4
        self.assertEqual(prompt.result_value, "")

    def test_result_value_multi(self):
        prompt = FuzzyPrompt(
            message="Select one of them", choices=[1, 2, 3], multiselect=True
        )
        self.assertEqual(prompt.result_value, [])
        prompt.content_control.choices[0]["enabled"] = True
        prompt.content_control.choices[2]["enabled"] = True
        self.assertEqual(prompt.result_value, [1, 3])

    def test_application(self):
        prompt = FuzzyPrompt(message="Select one of them", choices=lambda _: [1, 2, 3])
        prompt._application = None  # type: ignore
        try:
            prompt.application
        except NotImplementedError:
            pass
        else:
            self.fail("NotImplemented not raised.")

    def test_handle_down(self):
        prompt = FuzzyPrompt(
            message="Select one of them",
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 4)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)

    def test_handle_down_no_cycle(self):
        prompt = FuzzyPrompt(
            message="Select one of them",
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 4)
        prompt._handle_down(None)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 4)

    def test_handle_up(self):
        prompt = FuzzyPrompt(
            message="Select one of them",
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 4)
        prompt._handle_up(None)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)
        prompt._handle_up(None)
        prompt._handle_up(None)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 4)

    def test_handle_up_no_cycle(self):
        prompt = FuzzyPrompt(
            message="Select one of them",
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)
        prompt._handle_up(None)
        prompt._handle_up(None)
        prompt._handle_up(None)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)

        prompt = FuzzyPrompt(
            message="Select one of them",
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
            cycle=False,
        )

    @patch("InquirerPy.base.complex.shutil.get_terminal_size")
    @patch("InquirerPy.utils.shutil.get_terminal_size")
    def test_extra_lines_with_long_instruction(self, mocked_term, mocked_term2):
        mocked_term.return_value = (24, 80)
        mocked_term2.return_value = (24, 80)
        message = 15 * "i"
        qmark = "[?]"
        instruction = 3 * "i"
        long_instruction = 24 * "i"
        prompt = FuzzyPrompt(
            message=message,
            qmark=qmark,
            instruction=instruction,
            long_instruction=long_instruction,
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(prompt.extra_long_instruction_line_count, 0)

        long_instruction = 25 * "i"
        prompt = FuzzyPrompt(
            message=message,
            qmark=qmark,
            instruction=instruction,
            long_instruction=long_instruction,
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(prompt.extra_long_instruction_line_count, 1)

    @patch("InquirerPy.base.complex.shutil.get_terminal_size")
    @patch("InquirerPy.utils.shutil.get_terminal_size")
    def test_extra_lines_due_to_offset(self, mocked_term, mocked_term2):
        mocked_term.return_value = (24, 80)
        mocked_term2.return_value = (24, 80)
        message = 15 * "i"
        qmark = "[?]"
        instruction = 3 * "i"
        prompt = FuzzyPrompt(
            message=message,
            qmark=qmark,
            instruction=instruction,
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(
            prompt.extra_line_count,
            (len(qmark) + 1 + len(message) + 1 + len(instruction) + 1 - 1) // 24,
        )

        instruction = 4 * "i"
        prompt = FuzzyPrompt(
            message=message,
            qmark=qmark,
            instruction=instruction,
            choices=[
                "haah",
                "haha",
                "what",
                "waht",
                {"name": "weaht", "value": "weaht", "enabled": True},
            ],
        )
        self.assertEqual(
            prompt.extra_line_count,
            (len(qmark) + 1 + len(message) + 1 + len(instruction) + 1 - 1) // 24,
        )

    @patch("InquirerPy.base.complex.shutil.get_terminal_size")
    def test_height_offset(self, mocked_term) -> None:
        mocked_term.return_value = (24, 80)
        message = 15 * "i"
        qmark = "[?]"
        instruction = 4 * "i"
        prompt = FuzzyPrompt(
            message=message, qmark=qmark, instruction=instruction, choices=[1, 2, 3]
        )
        self.assertEqual(prompt.height_offset, 4)
        prompt._wrap_lines = False
        self.assertEqual(prompt.height_offset, 3)

        prompt = FuzzyPrompt(
            message=message,
            qmark=qmark,
            instruction=instruction,
            choices=[1, 2, 3],
            border=True,
        )
        self.assertEqual(prompt.height_offset, 6)

        instruction = 3 * "i"  # width total should be 24 and no extra lines needed
        prompt = FuzzyPrompt(
            message=message, qmark=qmark, instruction=instruction, choices=[1, 2, 3]
        )
        self.assertEqual(prompt.height_offset, 3)

    def test_get_error_message(self):
        class SelectionValidator(Validator):
            def validate(self, document) -> None:
                if document.text == 1:
                    raise ValidationError(
                        message="hello", cursor_position=document.cursor_position
                    )

        prompt = FuzzyPrompt(
            message="",
            choices=[1, 2, 3],
            validate=SelectionValidator(),
            invalid_message="minimum 2 selections",
        )
        self.assertEqual(prompt._invalid_message, "minimum 2 selections")
        self.assertFalse(prompt._invalid)
        prompt._handle_enter(None)
        self.assertEqual(prompt._invalid_message, "hello")
        self.assertTrue(prompt._invalid)
        self.assertEqual(
            prompt._get_error_message(), [("class:validation-toolbar", "hello")]
        )



from InquirerPy import inquirer
from InquirerPy.prompts.expand import ExpandChoice
from InquirerPy.separator import Separator

question1_choice = [
    ExpandChoice(key="a", name="Apple", value="Apple"),
    ExpandChoice(key="c", name="Cherry", value="Cherry"),
    ExpandChoice(key="o", name="Orange", value="Orange"),
    ExpandChoice(key="p", name="Peach", value="Peach"),
    ExpandChoice(key="m", name="Melon", value="Melon"),
    ExpandChoice(key="s", name="Strawberry", value="Strawberry"),
    ExpandChoice(key="g", name="Grapes", value="Grapes"),
]


def question2_choice(_):
    return [
        ExpandChoice(key="d", name="Delivery", value="dl"),
        ExpandChoice(key="p", name="Pick Up", value="pk"),
        Separator(line=15 * "*"),
        ExpandChoice(key="c", name="Car Park", value="cp"),
        ExpandChoice(key="t", name="Third Party", value="tp"),
    ]


def main():
    fruit = inquirer.expand(
        message="Pick your favourite:", choices=question1_choice, default="o"
    ).execute()
    method = inquirer.expand(
        message="Select your preferred method:", choices=question2_choice
    ).execute()


if __name__ == "__main__":
    main()



"""Contains the base class :class:`.BaseSimplePrompt`."""
import os
import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters.base import Condition, FilterOrBool
from prompt_toolkit.key_binding.key_bindings import KeyBindings, KeyHandlerCallable
from prompt_toolkit.keys import Keys
from prompt_toolkit.styles.style import Style
from prompt_toolkit.validation import Validator

from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT
from InquirerPy.exceptions import RequiredKeyNotFound
from InquirerPy.utils import (
    InquirerPyMessage,
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
    get_style,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent


class BaseSimplePrompt(ABC):
    """The base class to create a simple terminal input prompt.

    Note:
        No actual :class:`~prompt_toolkit.application.Application` is created by this class.
        This class only creates some common interface and attributes that can be easily used
        by `prompt_toolkit`.

        To have a functional prompt, you'll at least have to implement the :meth:`.BaseSimplePrompt._run`
        and :meth:`.BaseSimplePrompt._get_prompt_message`.

    See Also:
        :class:`~InquirerPy.prompts.input.InputPrompt`
    """

    def __init__(
        self,
        message: InquirerPyMessage,
        style: Optional[InquirerPyStyle] = None,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        instruction: str = "",
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        default: Any = "",
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        self._mandatory = mandatory
        self._mandatory_message = mandatory_message
        self._result = session_result or {}
        self._message = (
            message
            if not isinstance(message, Callable)
            else cast(Callable, message)(self._result)
        )
        self._instruction = instruction
        self._default = (
            default if not isinstance(default, Callable) else default(self._result)
        )
        self._style = Style.from_dict(style.dict if style else get_style().dict)
        self._qmark = qmark
        self._amark = amark
        self._status = {"answered": False, "result": None, "skipped": False}
        self._kb = KeyBindings()
        self._lexer = "class:input"
        self._transformer = transformer
        self._filter = filter
        self._wrap_lines = wrap_lines
        self._editing_mode = (
            EditingMode.VI
            if vi_mode or bool(os.getenv("INQUIRERPY_VI_MODE", False))
            else EditingMode.EMACS
        )
        if isinstance(validate, Validator):
            self._validator = validate
        else:
            self._validator = Validator.from_callable(
                validate if validate else lambda _: True,
                invalid_message,
                move_cursor_to_end=True,
            )
        self._raise_kbi = not os.getenv(
            "INQUIRERPY_NO_RAISE_KBI", not raise_keyboard_interrupt
        )
        self._is_rasing_kbi = Condition(lambda: self._raise_kbi)

        self._kb_maps = {
            "answer": [{"key": Keys.Enter}],
            "interrupt": [
                {"key": "c-c", "filter": self._is_rasing_kbi},
                {"key": "c-d", "filter": ~self._is_rasing_kbi},
            ],
            "skip": [{"key": "c-z"}, {"key": "c-c", "filter": ~self._is_rasing_kbi}],
        }
        self._kb_func_lookup = {
            "answer": [{"func": self._handle_enter}],
            "interrupt": [{"func": self._handle_interrupt}],
            "skip": [{"func": self._handle_skip}],
        }

    def _keybinding_factory(self):
        """Register all keybindings in `self._kb_maps`.

        It's required to call this function at the end of prompt constructor if
        it inherits from :class:`~InquirerPy.base.simple.BaseSimplePrompt` or
        :class:`~InquirerPy.base.complex.BaseComplexPrompt`.
        """

        def _factory(keys, filter, action):
            if action not in self.kb_func_lookup:
                raise RequiredKeyNotFound(f"keybinding action {action} not found")
            if not isinstance(keys, list):
                keys = [keys]

            @self.register_kb(*keys, filter=filter)
            def _(event):
                for method in self.kb_func_lookup[action]:
                    method["func"](event, *method.get("args", []))

        for key, item in self.kb_maps.items():
            if not isinstance(item, list):
                item = [item]
            for kb in item:
                _factory(kb["key"], kb.get("filter", Condition(lambda: True)), key)

    @abstractmethod
    def _set_error(self, message: str) -> None:
        """Set the error message for the prompt.

        Args:
            message: Error message to set.
        """
        pass

    def _handle_skip(self, event: Optional["KeyPressEvent"]) -> None:
        """Handle the event when attempting to skip a prompt.

        Skip the prompt if the `_mandatory` field is False, otherwise
        show an error message that the prompt cannot be skipped.
        """
        if not self._mandatory:
            self.status["answered"] = True
            self.status["skipped"] = True
            self.status["result"] = None
            if event:
                event.app.exit(result=None)
        else:
            self._set_error(message=self._mandatory_message)

    def _handle_interrupt(self, event: Optional["KeyPressEvent"]) -> None:
        """Handle the event when a KeyboardInterrupt signal is sent."""
        self.status["answered"] = True
        self.status["result"] = INQUIRERPY_KEYBOARD_INTERRUPT
        self.status["skipped"] = True
        if event:
            event.app.exit(result=INQUIRERPY_KEYBOARD_INTERRUPT)

    @abstractmethod
    def _handle_enter(self, event: Optional["KeyPressEvent"]) -> None:
        """Handle the event when user attempt to answer the question."""
        pass

    @property
    def status(self) -> Dict[str, Any]:
        """Dict[str, Any]: Get current prompt status.

        The status contains 3 keys: "answered" and "result".
            answered: If the current prompt is answered.
            result: The result of the user answer.
            skipped: If the prompt is skipped.
        """
        return self._status

    @status.setter
    def status(self, value) -> None:
        self._status = value

    def register_kb(
        self, *keys: Union[Keys, str], filter: FilterOrBool = True, **kwargs
    ) -> Callable[[KeyHandlerCallable], KeyHandlerCallable]:
        """Keybinding registration decorator.

        This decorator wraps around the :meth:`prompt_toolkit.key_binding.KeyBindings.add` with
        added feature to process `alt` realted keybindings.

        By default, `prompt_toolkit` doesn't process `alt` related keybindings,
        it requires `alt-ANY` to `escape` + `ANY`.

        Args:
            keys: The keys to bind that can trigger the function.
            filter: :class:`~prompt_toolkit.filter.Condition` to indicate if this keybinding should be active.

        Returns:
            A decorator that should be applied to the function thats intended to be active when the keys
            are pressed.

        Examples:
            >>> @self.register_kb("alt-j")
            ... def test(event):
            ...     pass
        """
        alt_pattern = re.compile(r"^alt-(.*)")

        def decorator(func: KeyHandlerCallable) -> KeyHandlerCallable:
            formatted_keys = []
            for key in keys:
                match = alt_pattern.match(key)
                if match:
                    formatted_keys.append("escape")
                    formatted_keys.append(match.group(1))
                else:
                    formatted_keys.append(key)

            @self._kb.add(*formatted_keys, filter=filter, **kwargs)
            def executable(event) -> None:
                func(event)

            return executable

        return decorator

    @abstractmethod
    def _get_prompt_message(
        self, pre_answer: Tuple[str, str], post_answer: Tuple[str, str]
    ) -> List[Tuple[str, str]]:
        """Get the question message in formatted text form to display in the prompt.

        This function is mainly used to render the question message dynamically based
        on the current status (answered or not answered) of the prompt.

        Note:
            The function requires implementation when inheriting :class:`.BaseSimplePrompt`.
            You should call `super()._get_prompt_message(pre_answer, post_answer)` in
            the implemented `_get_prompt_message`.

        Args:
            pre_answer: The message to display before the question is answered.
            post_answer: The information to display after the question is answered.

        Returns:
            Formatted text in list of tuple format.
        """
        display_message = []
        if self.status["skipped"]:
            display_message.append(("class:skipped", self._qmark))
            display_message.append(
                ("class:skipped", "%s%s " % (" " if self._qmark else "", self._message))
            )
        elif self.status["answered"]:
            display_message.append(("class:answermark", self._amark))
            display_message.append(
                (
                    "class:answered_question",
                    "%s%s" % (" " if self._amark else "", self._message),
                )
            )
            display_message.append(
                post_answer
                if not self._transformer
                else (
                    "class:answer",
                    " %s" % self._transformer(self.status["result"]),
                )
            )
        else:
            display_message.append(("class:questionmark", self._qmark))
            display_message.append(
                (
                    "class:question",
                    "%s%s" % (" " if self._qmark else "", self._message),
                )
            )
            display_message.append(pre_answer)
        return display_message

    @abstractmethod
    def _run(self) -> Any:
        """Abstractmethod to enforce a run function is implemented.

        All prompt instance requires a `_run` call to initialise and run an instance of
        `PromptSession` or `Application`.
        """
        pass

    @abstractmethod
    async def _run_async(self) -> Any:
        """Abstractmethod to enforce a run function is implemented.

        All prompt instance requires a `_run_async` call to initialise and run an instance of
        `PromptSession` or `Application`.
        """
        pass

    def execute(self, raise_keyboard_interrupt: Optional[bool] = None) -> Any:
        """Run the prompt and get the result.

        Args:
            raise_keyboard_interrupt: **Deprecated**. Set this parameter on the prompt initialisation instead.

        Returns:
            Value of the user answer. Types varies depending on the prompt.

        Raises:
            KeyboardInterrupt: When `ctrl-c` is pressed and `raise_keyboard_interrupt` is True.
        """
        result = self._run()
        if raise_keyboard_interrupt is not None:
            self._raise_kbi = not os.getenv(
                "INQUIRERPY_NO_RAISE_KBI", not raise_keyboard_interrupt
            )
        if result == INQUIRERPY_KEYBOARD_INTERRUPT:
            raise KeyboardInterrupt
        if not self._filter:
            return result
        return self._filter(result)

    async def execute_async(self) -> None:
        """Run the prompt asynchronously and get the result.

        Returns:
            Value of the user answer. Types varies depending on the prompt.

        Raises:
            KeyboardInterrupt: When `ctrl-c` is pressed and `raise_keyboard_interrupt` is True.
        """
        result = await self._run_async()
        if result == INQUIRERPY_KEYBOARD_INTERRUPT:
            raise KeyboardInterrupt
        if not self._filter:
            return result
        return self._filter(result)

    @property
    def instruction(self) -> str:
        """str: Instruction to display next to question."""
        return self._instruction

    @property
    def kb_maps(self) -> Dict[str, Any]:
        """Dict[str, Any]: Keybinding mappings."""
        return self._kb_maps

    @kb_maps.setter
    def kb_maps(self, value: Dict[str, Any]) -> None:
        self._kb_maps = {**self._kb_maps, **value}

    @property
    def kb_func_lookup(self) -> Dict[str, Any]:
        """Dict[str, Any]: Keybinding function lookup mappings.."""
        return self._kb_func_lookup

    @kb_func_lookup.setter
    def kb_func_lookup(self, value: Dict[str, Any]) -> None:
        self._kb_func_lookup = {**self._kb_func_lookup, **value}



import unittest
from unittest.mock import patch

from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT, INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.prompts.list import InquirerPyListControl, ListPrompt
from InquirerPy.separator import Separator
from InquirerPy.utils import InquirerPyStyle


class TestListPrompt(unittest.TestCase):
    choices = [
        {"name": "apple", "value": "peach"},
        "pear",
        {"name": "melon", "value": "watermelon"},
    ]

    def test_list_control_enabled(self) -> None:
        list_control = InquirerPyListControl(
            [
                {"name": "apple", "value": "peach", "enabled": True},
                "pear",
                {"name": "melon", "value": "watermelon"},
            ],
            "watermelon",
            INQUIRERPY_POINTER_SEQUENCE,
            ">",
            None,
            True,
            " ",
        )
        self.assertEqual(
            list_control.choices,
            [
                {"name": "apple", "value": "peach", "enabled": True},
                {"name": "pear", "value": "pear", "enabled": False},
                {"name": "melon", "value": "watermelon", "enabled": False},
            ],
        )

    def test_list_control_exceptions(self):
        self.assertRaises(
            RequiredKeyNotFound,
            InquirerPyListControl,
            [
                {"what": "apple", "value": "peach"},
                "pear",
            ],
            "watermelon",
            "",
            "",
            None,
            False,
            " ",
        )
        self.assertRaises(
            InvalidArgument, InquirerPyListControl, [], "", "", "", None, False, " "
        )

    def test_choice_combination(self):
        prompt = ListPrompt(message="Test combo", choices=["hello"])
        self.assertEqual(prompt._qmark, "?")
        self.assertEqual(prompt.instruction, "")

        self.assertRaises(InvalidArgument, ListPrompt, "", [Separator(), Separator()])

    def test_list_prompt_message(self):
        prompt = ListPrompt(
            message="Select a fruit",
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark="[?]",
            pointer=">",
            instruction="(j/k)",
        )
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:questionmark", "[?]"),
                ("class:question", " Select a fruit"),
                ("class:instruction", " (j/k) "),
            ],
        )

    def test_list_bindings(self):
        prompt = ListPrompt(
            message="Select a fruit",
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark="[?]",
            pointer=">",
            instruction="(j/k)",
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 2)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)

        self.assertEqual(
            prompt.status, {"result": None, "answered": False, "skipped": False}
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(
            prompt.status, {"result": "melon", "answered": True, "skipped": False}
        )

    def test_separator_movement(self):
        prompt = ListPrompt(message="..", choices=[Separator("hello"), "yes"])
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)

        prompt = ListPrompt(
            message="..", choices=[Separator("hello"), "yes", Separator(), "no"]
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)

    def test_list_enter_empty(self):
        prompt = ListPrompt(
            message="",
            choices=["haha", "haah", "what", "I don't know"],
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
            self.assertEqual(prompt.status["result"], "haha")

        prompt = ListPrompt(
            message="",
            choices=["haha", "haah", "what", "I don't know"],
            multiselect=True,
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
            self.assertEqual(prompt.status["result"], ["haha"])
            prompt.content_control.choices[1]["enabled"] = True
            prompt._handle_enter(event)
            self.assertEqual(prompt.status["result"], ["haah"])

    @patch("InquirerPy.base.complex.Application.run")
    def test_prompt_execute(self, mocked_run):
        mocked_run.return_value = "hello"
        result = ListPrompt("hello world", ["yes", "no"]).execute()
        self.assertEqual(result, "hello")

        result = ListPrompt(
            "hello world", ["yes", "no"], filter=lambda _: "no"
        ).execute()
        self.assertEqual(result, "no")

        mocked_run.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        prompt = ListPrompt("hello world", ["yes", "no"], filter=lambda _: "no")
        self.assertRaises(KeyboardInterrupt, prompt.execute)

        mocked_run.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        prompt = ListPrompt("hello world", ["yes", "no"])
        try:
            result = prompt.execute(raise_keyboard_interrupt=False)
        except KeyboardInterrupt:
            self.assertFalse(prompt._is_rasing_kbi())
        else:
            self.fail("should raise kbi")



import asyncio
import unittest

from prompt_toolkit.filters.base import Condition, Filter
from prompt_toolkit.layout.containers import ConditionalContainer

from InquirerPy.containers.spinner import SPINNERS, SpinnerWindow


class TestSpinner(unittest.TestCase):
    def setUp(self) -> None:
        self.spinner = SpinnerWindow(Condition(lambda: True), redraw=lambda: None)

    def test_init(self) -> None:
        self.assertIsInstance(self.spinner, ConditionalContainer)
        self.assertIsInstance(self.spinner._loading, Filter)
        self.assertEqual(self.spinner._pattern, SPINNERS.line)
        self.assertEqual(self.spinner._text, "Loading ...")

        spinner = SpinnerWindow(
            Condition(lambda: True),
            redraw=lambda: None,
            delay=0.2,
            pattern=SPINNERS.dots,
            text="Hello",
        )
        self.assertEqual(spinner._pattern, SPINNERS.dots)
        self.assertEqual(spinner._text, "Hello")
        self.assertEqual(spinner._delay, 0.2)

    def test_get_text(self):
        self.assertEqual(
            self.spinner._get_text(),
            [
                ("class:spinner_pattern", SPINNERS.line[0]),
                ("", " "),
                ("class:spinner_text", "Loading ..."),
            ],
        )

    def test_start(self):
        flag = {"loading": True, "counter": 0}

        async def run_start(spinner) -> None:
            asyncio.create_task(spinner.start())
            asyncio.create_task(spinner.start())
            await asyncio.sleep(0.4)
            flag["loading"] = False

        def count():
            flag["counter"] += 1

        spinner = SpinnerWindow(Condition(lambda: flag["loading"]), redraw=count)
        asyncio.run(run_start(spinner))
        self.assertNotEqual(flag["counter"], 0)



import unittest

from InquirerPy.containers.instruction import InstructionWindow


class TestInstructionWindow(unittest.TestCase):
    def test_get_message(self):
        window = InstructionWindow(message="hello", filter=True, wrap_lines=True)
        self.assertEqual(window._get_message(), [("class:long_instruction", "hello")])



from InquirerPy import inquirer
from InquirerPy.validator import NumberValidator

age = inquirer.text(
    message="Enter your age:",
    validate=NumberValidator(),
    default="18",
    filter=lambda result: int(result),
    transformer=lambda result: "Adult" if int(result) >= 18 else "Youth",
).execute()

drinks = ["Soda", "Cidr", "Water", "Milk"] if age < 18 else ["Wine", "Beer"]

drink = inquirer.rawlist(
    message="What drinks would you like to buy:", default=2, choices=drinks
).execute()

if drink in {"Wine", "Beer"}:
    bag = inquirer.select(
        message="Would you like a bag:", choices=["Yes", "No"]
    ).execute()

confirm = inquirer.confirm(message="Confirm?", default=True).execute()



import asyncio
import unittest

from InquirerPy.base.control import Choice
from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.prompts.list import InquirerPyListControl
from InquirerPy.separator import Separator


class TestBaseControl(unittest.TestCase):
    def test_constructor(self) -> None:
        control = InquirerPyListControl(
            choices=[1, 2, 3],
            default=None,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control._default, None)
        self.assertEqual(
            control.choices,
            [
                {"name": "1", "value": 1, "enabled": False},
                {"name": "2", "value": 2, "enabled": False},
                {"name": "3", "value": 3, "enabled": False},
            ],
        )

    def test_constructor_choice_callable(self):
        control = InquirerPyListControl(
            choices=lambda _: [1, 2, 3],
            default=lambda _: 1,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control._default, 1)
        self.assertEqual(
            control.choices,
            [
                {"name": "1", "value": 1, "enabled": False},
                {"name": "2", "value": 2, "enabled": False},
                {"name": "3", "value": 3, "enabled": False},
            ],
        )

    def test_retrieve_choices(self) -> None:
        control = InquirerPyListControl(
            choices=lambda _: [1, 2, 3],
            default=lambda _: 1,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(
            control.choices,
            [
                {"name": "1", "value": 1, "enabled": False},
                {"name": "2", "value": 2, "enabled": False},
                {"name": "3", "value": 3, "enabled": False},
            ],
        )

    def test_get_choices_list(self):
        sep = Separator()
        choice = [1, 2, 3, sep]
        default = 2
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control.selected_choice_index, 1)
        choices = control._get_choices(choice, 1)
        self.assertEqual(
            choices,
            [
                {"name": "1", "value": 1, "enabled": False},
                {"name": "2", "value": 2, "enabled": False},
                {"name": "3", "value": 3, "enabled": False},
                {"enabled": False, "name": "---------------", "value": sep},
            ],
        )
        self.assertEqual(control.selected_choice_index, 0)

    def test_get_choices_dict(self):
        sep = Separator()
        choice = [
            {"name": "1", "value": 1, "enabled": True},
            {"name": "2", "value": 2, "enabled": True},
            {"name": "3", "value": 3, "enabled": True},
            sep,
        ]
        default = 2
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control.selected_choice_index, 1)
        choices = control._get_choices(choice, 1)
        self.assertEqual(
            choices,
            [
                {"name": "1", "value": 1, "enabled": False},
                {"name": "2", "value": 2, "enabled": False},
                {"name": "3", "value": 3, "enabled": False},
                {"enabled": False, "name": "---------------", "value": sep},
            ],
        )
        self.assertEqual(control.selected_choice_index, 0)

        control._multiselect = True
        choices = control._get_choices(choice, 1)
        self.assertEqual(
            choices,
            [
                {"name": "1", "value": 1, "enabled": True},
                {"name": "2", "value": 2, "enabled": True},
                {"name": "3", "value": 3, "enabled": True},
                {"enabled": False, "name": "---------------", "value": sep},
            ],
        )

    def test_get_choices_key_error(self):
        choice = [
            {"asdf": "1", "value": 1, "enabled": True},
            {"name": "2", "value": 2, "enabled": True},
            {"name": "3", "value": 3, "enabled": True},
        ]
        control = InquirerPyListControl(
            choices=[1],
            default=None,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertRaises(RequiredKeyNotFound, control._get_choices, choice, None)

    def test_safety_check(self):
        control = InquirerPyListControl(
            choices=[1],
            default=None,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        control._choices = []
        self.assertRaises(InvalidArgument, control._safety_check)

        control._choices = control._get_choices([Separator(), Separator()], None)
        self.assertRaises(InvalidArgument, control._safety_check)

    def test_get_formatted_choices(self):
        list_control = InquirerPyListControl(
            [
                {"name": "apple", "value": "peach"},
                "pear",
                {"name": "melon", "value": "watermelon"},
            ],
            "watermelon",
            INQUIRERPY_POINTER_SEQUENCE,
            ">",
            None,
            True,
            " ",
        )
        self.assertEqual(
            list_control._get_formatted_choices(),
            [
                ("", " "),
                ("class:marker", " "),
                ("", "apple"),
                ("", "\n"),
                ("", " "),
                ("class:marker", " "),
                ("", "pear"),
                ("", "\n"),
                ("class:pointer", "❯"),
                ("class:marker", " "),
                ("[SetCursorPosition]", ""),
                ("class:pointer", "melon"),
            ],
        )
        list_control.choices[0]["enabled"] = True
        list_control.choices[1]["enabled"] = True
        self.assertEqual(
            list_control._get_formatted_choices(),
            [
                ("", " "),
                ("class:marker", ">"),
                ("", "apple"),
                ("", "\n"),
                ("", " "),
                ("class:marker", ">"),
                ("", "pear"),
                ("", "\n"),
                ("class:pointer", "❯"),
                ("class:marker", " "),
                ("[SetCursorPosition]", ""),
                ("class:pointer", "melon"),
            ],
        )

    def test_choice_count(self):
        choice = [
            {"name": "1", "value": 1, "enabled": True},
            {"name": "2", "value": 2, "enabled": True},
            {"name": "3", "value": 3, "enabled": True},
        ]
        default = 2
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control.choice_count, 3)

    def test_selection(self):
        choice = [
            {"name": "1", "value": 1, "enabled": True},
            {"name": "2", "value": 2, "enabled": True},
            {"name": "3", "value": 3, "enabled": True},
        ]
        default = 2
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control.selection, {"name": "2", "value": 2, "enabled": False})

    def test_choice(self):
        choice = [Choice(1), Choice(2), Choice(3, None, True)]

        default = 2
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=False,
            marker_pl=" ",
        )
        self.assertEqual(control.selection, {"name": "2", "value": 2, "enabled": False})
        self.assertEqual(
            control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": False, "name": "3", "value": 3},
            ],
        )

    def test_choice_multi(self):
        choice = [Choice(1), Choice(2), Choice(3, None, True)]

        default = 4
        control = InquirerPyListControl(
            choices=choice,
            default=default,
            pointer=INQUIRERPY_POINTER_SEQUENCE,
            marker=INQUIRERPY_POINTER_SEQUENCE,
            session_result=None,
            multiselect=True,
            marker_pl=" ",
        )
        self.assertEqual(control.selection, {"name": "1", "value": 1, "enabled": False})
        self.assertEqual(
            control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": True, "name": "3", "value": 3},
            ],
        )



import urllib.request
from contextlib import ExitStack
from pathlib import Path

from InquirerPy import inquirer


def get_choices(_):
    p = Path(__file__).resolve().parent.joinpath("sample.txt")
    choices = []
    result = []

    with ExitStack() as stack:
        if not p.exists():
            file = stack.enter_context(p.open("w+"))
            sample = stack.enter_context(
                urllib.request.urlopen(
                    "https://assets.kazhala.me/InquirerPy/sample.txt"
                )
            )
            file.write(sample.read().decode())
            file.seek(0, 0)
        else:
            file = stack.enter_context(p.open("r"))
        for line in file.readlines():
            choices.append(line[:-1])
        for choice in choices:
            if not choice:
                continue
            result.append(choice)
    return result


def main():
    action = inquirer.fuzzy(
        message="Select actions:",
        choices=["hello", "weather", "what", "whoa", "hey", "yo"],
        default="he",
    ).execute()
    words = inquirer.fuzzy(
        message="Select preferred words:",
        choices=get_choices,
        multiselect=True,
        validate=lambda result: len(result) > 1,
        invalid_message="minimum 2 selections",
        max_height="70%",
    ).execute()


if __name__ == "__main__":
    main()



import asyncio

from InquirerPy import inquirer, prompt_async


async def main():
    questions = [
        {"type": "input", "message": "Name:"},
        {"type": "number", "message": "Number:"},
        {"type": "confirm", "message": "Confirm?"},
    ]
    result = await prompt_async(questions)
    name = await inquirer.text(message="Name:").execute_async()
    number = await inquirer.number(message="Number:").execute_async()
    confirm = await inquirer.confirm(message="Confirm?").execute_async()


if __name__ == "__main__":
    asyncio.run(main())



"""Module contains the main message window :class:`~prompt_toolkit.container.Container`."""

from typing import TYPE_CHECKING

from prompt_toolkit.layout.containers import ConditionalContainer, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension

if TYPE_CHECKING:
    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text.base import AnyFormattedText


class MessageWindow(ConditionalContainer):
    """Main window to display question to the user.

    Args:
        message: The message to display in the terminal.
        filter: Condition that this message window should be displayed.
            Use a loading condition to only display this window while its not loading.
        wrap_lines: Enable line wrapping if the message is too long.
        show_cursor: Display cursor.
    """

    def __init__(
        self,
        message: "AnyFormattedText",
        filter: "FilterOrBool",
        wrap_lines: bool = True,
        show_cursor: bool = True,
        **kwargs
    ) -> None:
        super().__init__(
            content=Window(
                height=LayoutDimension.exact(1) if not wrap_lines else None,
                content=FormattedTextControl(message, show_cursor=show_cursor),
                wrap_lines=wrap_lines,
                dont_extend_height=True,
                **kwargs
            ),
            filter=filter,
        )



import os
import shutil
import tempfile
import unittest
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import ANY, call, patch

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import CompleteEvent
from prompt_toolkit.document import Document
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.input.defaults import create_pipe_input
from prompt_toolkit.output import DummyOutput
from prompt_toolkit.shortcuts.prompt import CompleteStyle

from InquirerPy.exceptions import InvalidArgument
from InquirerPy.prompts.filepath import FilePathCompleter, FilePathPrompt
from InquirerPy.utils import InquirerPyStyle
from InquirerPy.validator import PathValidator


class TestFilePath(unittest.TestCase):
    def setUp(self):
        self.inp = create_pipe_input()
        self.dirs_to_create = ["dir1", "dir2", "dir3", ".dir"]
        self.files_to_create = ["file1", "file2", "file3", ".file"]
        self.test_dir = Path(tempfile.mkdtemp())
        self.create_temp_files()

    def tearDown(self):
        self.inp.close()
        shutil.rmtree(self.test_dir)

    @contextmanager
    def chdir(self, directory):
        orig_dir = os.getcwd()
        os.chdir(directory)
        try:
            yield
        finally:
            os.chdir(orig_dir)

    def create_temp_files(self):
        for directory in self.dirs_to_create:
            self.test_dir.joinpath(directory).mkdir(exist_ok=True)
        for file in self.files_to_create:
            with self.test_dir.joinpath(file).open("wb") as output_file:
                output_file.write("".encode("UTF-8"))

    def test_completer_explicit_currdir_all(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = "./"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(
                sorted(completions),
                sorted(self.dirs_to_create + self.files_to_create),
            )

    def test_completer_currdir_file(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = "./file"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), ["file1", "file2", "file3"])

    def test_completer_hidden(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = "."
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), [".dir", ".file"])

    def test_completer_normal(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = "dir"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), ["dir1", "dir2", "dir3"])

    def test_completer_expanduser(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = "~/"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertGreater(len(completions), 0)

    def test_completer_dir_only(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter(only_directories=True)
            doc_text = "./"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), sorted(self.dirs_to_create))

    def test_completer_file_only(self):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter(only_files=True)
            doc_text = "./"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), sorted(self.files_to_create))

    def test_input(self):
        self.inp.send_text("./file1\n")
        filepath_prompt = FilePathPrompt(
            message="hello",
            style=InquirerPyStyle({"qmark": "bold"}),
            input=self.inp,
            output=DummyOutput(),
        )
        result = filepath_prompt.execute()
        self.assertEqual(result, "./file1")
        self.assertEqual(filepath_prompt.status["answered"], True)
        self.assertEqual(filepath_prompt.status["result"], "./file1")

    def test_default_answer(self):
        self.inp.send_text("\n")
        filepath_prompt = FilePathPrompt(
            message="hello",
            style=InquirerPyStyle({"qmark": "bold"}),
            default=".vim",
            input=self.inp,
            output=DummyOutput(),
        )
        result = filepath_prompt.execute()
        self.assertEqual(result, ".vim")
        self.assertEqual(filepath_prompt.status["answered"], True)
        self.assertEqual(filepath_prompt.status["result"], ".vim")

    @patch.object(Buffer, "validate_and_handle")
    def test_validation(self, mocked_validate):
        def _hello():
            filepath_prompt._session.app.exit(result="hello")

        mocked_validate.side_effect = _hello
        self.inp.send_text("hello\n")
        filepath_prompt = FilePathPrompt(
            message="fooboo",
            style=InquirerPyStyle({"qmark": ""}),
            default=".vim",
            validate=PathValidator(),
            input=self.inp,
            output=DummyOutput(),
        )
        result = filepath_prompt.execute()
        mocked_validate.assert_called_once()
        self.assertEqual(result, "hello")
        self.assertEqual(filepath_prompt.status["answered"], False)
        self.assertEqual(filepath_prompt.status["result"], None)

    def test_get_prompt_message(self):
        filepath_prompt = FilePathPrompt(
            message="brah", style=InquirerPyStyle({"foo": ""}), qmark="!", amark="x"
        )
        message = filepath_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "!"),
                ("class:question", " brah"),
                ("class:instruction", " "),
            ],
        )

        filepath_prompt.status["answered"] = True
        filepath_prompt.status["result"] = "hello"
        message = filepath_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "x"),
                ("class:answered_question", " brah"),
                ("class:answer", " hello"),
            ],
        )

    @patch("InquirerPy.prompts.input.SimpleLexer")
    @patch("InquirerPy.prompts.filepath.FilePathPrompt._get_prompt_message")
    @patch("InquirerPy.base.simple.Style.from_dict")
    @patch("InquirerPy.base.simple.KeyBindings")
    @patch("InquirerPy.prompts.input.PromptSession")
    def test_callable_called(
        self,
        MockedSession,
        MockedKeyBindings,
        MockedStyle,
        mocked_message,
        MockedLexer,
    ):
        def _validation(_):
            return True

        FilePathPrompt(
            message="yes",
            style=InquirerPyStyle({"yes": ""}),
            default="",
            qmark="XD",
            multicolumn_complete=True,
            validate=_validation,
            vi_mode=True,
            only_directories=True,
        )
        kb = MockedKeyBindings()
        style = MockedStyle()
        lexer = MockedLexer()
        MockedSession.assert_called_once_with(
            message=mocked_message,
            key_bindings=kb,
            style=style,
            completer=ANY,
            validator=ANY,
            validate_while_typing=False,
            input=None,
            output=None,
            editing_mode=EditingMode.VI,
            lexer=lexer,
            is_password=False,
            multiline=False,
            complete_style=CompleteStyle.MULTI_COLUMN,
            wrap_lines=True,
            bottom_toolbar=None,
        )

        MockedStyle.assert_has_calls([call({"yes": ""})])

    def test_invalid_argument(self):
        self.assertRaises(InvalidArgument, FilePathPrompt, "hello", None, False, 12)
        FilePathPrompt(message="hello", default=lambda _: "12")

    @patch("os.name")
    def test_completer_explicit_currdir_all_win(self, mocked_platform):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = ".\\"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(
                sorted(completions),
                sorted(self.dirs_to_create + self.files_to_create),
            )

    @patch("os.name")
    def test_completer_currdir_file_win(self, mocked_platform):
        with self.chdir(self.test_dir):
            completer = FilePathCompleter()
            doc_text = ".\\file"
            doc = Document(doc_text, len(doc_text))
            event = CompleteEvent()
            completions = [
                completion.text
                for completion in list(completer.get_completions(doc, event))
            ]
            self.assertEqual(sorted(completions), ["file1", "file2", "file3"])



from InquirerPy import prompt
from InquirerPy.base.control import Choice
from InquirerPy.separator import Separator


def main():
    questions = [
        {
            "type": "list",
            "message": "Select an action:",
            "choices": ["Upload", "Download", Choice(value=None, name="Exit")],
            "default": None,
        },
        {
            "type": "list",
            "message": "Select regions:",
            "choices": [
                Choice("ap-southeast-2", name="Sydney"),
                Choice("ap-southeast-1", name="Singapore"),
                Separator(),
                "us-east-1",
                "us-east-2",
            ],
            "multiselect": True,
            "transformer": lambda result: f"{len(result)} region{'s' if len(result) > 1 else ''} selected",
            "when": lambda result: result[0] is not None,
        },
    ]

    result = prompt(questions=questions)


if __name__ == "__main__":
    main()



"""Contains the content control class :class:`.InquirerPyUIListControl`."""
from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

from prompt_toolkit.layout.controls import FormattedTextControl

from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.separator import Separator
from InquirerPy.utils import InquirerPyListChoices, InquirerPySessionResult

__all__ = ["Choice", "InquirerPyUIListControl"]


@dataclass
class Choice:
    """Class to create choices for list type prompts.

    A simple dataclass that can be used as an alternate to using :class:`dict`
    when working with choices.

    Args:
        value: The value of the choice when user selects this choice.
        name: The value that should be presented to the user prior/after selection of the choice.
            This value is optional, if not provided, it will fallback to the string representation of `value`.
        enabled: Indicates if the choice should be pre-selected.
            This only has effects when the prompt has `multiselect` enabled.
    """

    value: Any
    name: Optional[str] = None
    enabled: bool = False

    def __post_init__(self):
        """Assign strinify value to name if not present."""
        if self.name is None:
            self.name = str(self.value)


class InquirerPyUIListControl(FormattedTextControl):
    """A base class to create :class:`~prompt_toolkit.layout.UIControl` to display list type contents.

    Args:
        choices(InquirerPyListChoices): List of choices to display as the content.
            Can also be a callable or async callable that returns a list of choices.
        default: Default value, this will affect the cursor position.
        multiselect: Indicate if the current prompt has `multiselect` enabled.
        session_result: Current session result.
    """

    def __init__(
        self,
        choices: InquirerPyListChoices,
        default: Any = None,
        multiselect: bool = False,
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        self._session_result = session_result or {}
        self._selected_choice_index: int = 0
        self._choice_func = None
        self._multiselect = multiselect
        self._default = (
            default
            if not isinstance(default, Callable)
            else cast(Callable, default)(self._session_result)
        )
        self._raw_choices = (
            choices
            if not isinstance(choices, Callable)
            else cast(Callable, choices)(self._session_result)
        )
        self._choices = self._get_choices(self._raw_choices, self._default)
        self._safety_check()
        self._format_choices()
        super().__init__(self._get_formatted_choices)

    def _get_choices(self, choices: List[Any], default: Any) -> List[Dict[str, Any]]:
        """Process the raw user input choices and format it into dictionary.

        Args:
            choices: List of chices to display.
            default: Default value, this will affect the :attr:`.InquirerPyUIListControl.selected_choice_index`

        Returns:
            List of choices.

        Raises:
            RequiredKeyNotFound: When the provided choice is missing the `name` or `value` key.
        """
        processed_choices: List[Dict[str, Any]] = []
        try:
            for index, choice in enumerate(choices, start=0):
                if isinstance(choice, dict):
                    if choice["value"] == default:
                        self.selected_choice_index = index
                    processed_choices.append(
                        {
                            "name": str(choice["name"]),
                            "value": choice["value"],
                            "enabled": choice.get("enabled", False)
                            if self._multiselect
                            else False,
                        }
                    )
                elif isinstance(choice, Separator):
                    if self.selected_choice_index == index:
                        self.selected_choice_index = (
                            self.selected_choice_index + 1
                        ) % len(choices)
                    processed_choices.append(
                        {"name": str(choice), "value": choice, "enabled": False}
                    )
                elif isinstance(choice, Choice):
                    dict_choice = asdict(choice)
                    if dict_choice["value"] == default:
                        self.selected_choice_index = index
                    if not self._multiselect:
                        dict_choice["enabled"] = False
                    processed_choices.append(dict_choice)
                else:
                    if choice == default:
                        self.selected_choice_index = index
                    processed_choices.append(
                        {"name": str(choice), "value": choice, "enabled": False}
                    )
        except KeyError:
            raise RequiredKeyNotFound(
                "dictionary type of choice require a 'name' key and a 'value' key"
            )
        return processed_choices

    @property
    def selected_choice_index(self) -> int:
        """int: Current highlighted index."""
        return self._selected_choice_index

    @selected_choice_index.setter
    def selected_choice_index(self, value: int) -> None:
        self._selected_choice_index = value

    @property
    def choices(self) -> List[Dict[str, Any]]:
        """List[Dict[str, Any]]: Get all processed choices."""
        return self._choices

    @choices.setter
    def choices(self, value: List[Dict[str, Any]]) -> None:
        self._choices = value

    def _safety_check(self) -> None:
        """Validate processed choices.

        Check if the choices are empty or if it only contains :class:`~InquirerPy.separator.Separator`.
        """
        if not self.choices:
            raise InvalidArgument("argument choices cannot be empty")
        should_proceed: bool = False
        for choice in self.choices:
            if not isinstance(choice["value"], Separator):
                should_proceed = True
                break
        if not should_proceed:
            raise InvalidArgument(
                "argument choices should contain choices other than separator"
            )

    def _get_formatted_choices(self) -> List[Tuple[str, str]]:
        """Get all choices in formatted text format.

        Returns:
            List of choices in formatted text form.
        """
        display_choices = []

        for index, choice in enumerate(self.choices):
            if index == self.selected_choice_index:
                display_choices += self._get_hover_text(choice)
            else:
                display_choices += self._get_normal_text(choice)
            display_choices.append(("", "\n"))
        if display_choices:
            display_choices.pop()
        return display_choices

    def _format_choices(self) -> None:
        """Perform post processing on the choices.

        Additional customisation to the choices after :meth:`.InquirerPyUIListControl._get_choices` call.
        """
        pass

    @abstractmethod
    def _get_hover_text(self, choice) -> List[Tuple[str, str]]:
        """Generate the formatted text for hovered choice.

        Returns:
            Formatted text in list of tuple format.
        """
        pass

    @abstractmethod
    def _get_normal_text(self, choice) -> List[Tuple[str, str]]:
        """Generate the formatted text for non-hovered choices.

        Returns:
            Formatted text in list of tuple format.
        """
        pass

    @property
    def choice_count(self) -> int:
        """int: Total count of choices."""
        return len(self.choices)

    @property
    def selection(self) -> Dict[str, Any]:
        """Dict[str, Any]: Current selected choice."""
        return self.choices[self.selected_choice_index]

    @property
    def loading(self) -> bool:
        """bool: Indicate if the content control is loading."""
        return self._loading

    @loading.setter
    def loading(self, value: bool) -> None:
        self._loading = value



"""Module contains shared utility functions and typing aliases."""
import math
import os
import shutil
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from prompt_toolkit import print_formatted_text
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.application.current import get_app
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import Validator

from InquirerPy.exceptions import InvalidArgument

if TYPE_CHECKING:
    from prompt_toolkit.filters.base import FilterOrBool

    from InquirerPy.base.control import Choice

__all__ = [
    "get_style",
    "calculate_height",
    "InquirerPyStyle",
    "patched_print",
    "color_print",
]


class InquirerPyStyle(NamedTuple):
    """`InquirerPy` Style class.

    Used as a helper class to enforce the method `get_style` to be used
    while also avoiding :class:`dict` to be passed into prompts.

    Note:
        The class is an instance of :class:`typing.NamedTuple`.

    Warning:
        You should not directly be using this class besides for type hinting
        purposes. Obtain an instance of this class using :func:`.get_style`.
    """

    dict: Dict[str, str]


InquirerPySessionResult = Dict[Union[str, int], Optional[Union[str, bool, List[Any]]]]
InquirerPyChoice = Union[List[Any], List["Choice"], List[Dict[str, Any]]]
InquirerPyListChoices = Union[
    Callable[["InquirerPySessionResult"], InquirerPyChoice],
    InquirerPyChoice,
]
InquirerPyValidate = Union[Callable[[Any], bool], "Validator"]
InquirerPyQuestions = Union[List[Dict[str, Any]], Dict[str, Any]]
InquirerPyMessage = Union[str, Callable[["InquirerPySessionResult"], str]]
InquirerPyDefault = Union[Any, Callable[["InquirerPySessionResult"], Any]]
InquirerPyKeybindings = Dict[
    str, List[Dict[str, Union[str, "FilterOrBool", List[str]]]]
]


def get_style(
    style: Optional[Dict[str, str]] = None, style_override: bool = True
) -> InquirerPyStyle:
    """Obtain an :class:`.InquirerPyStyle` instance which can be consumed by the `style` parameter in prompts.

    Tip:
        This function supports ENV variables.

        For all the color ENV variable names, refer to the :ref:`ENV <pages/env:Style>` documentation.

    Note:
        If no style is provided, then a default theme based on `one dark <https://github.com/joshdick/onedark.vim#color-reference>`_
        color palette is applied.

    Note:
        Priority: style parameter -> ENV variable -> default style

    Args:
        style: The dictionary of style classes and their colors, If nothing is passed, the style will be resolved to the :ref:`pages/style:Default Style`.
        style_override: A boolean to determine if the supplied `style` parameter should be merged with the :ref:`pages/style:Default Style` or override them.
            By default, the supplied style will overwrite the :ref:`pages/style:Default Style`.

    Returns:
        An instance of :class:`.InquirerPyStyle`.

    Examples:
        >>> from InquirerPy import get_style
        >>> from InquirerPy import inquirer
        >>> style = get_style({"questionmark": "#ffffff", "answer": "#000000"}, style_override=False)
        >>> result = inquirer.confirm(message="Confirm?", style=style).execute()
    """
    if not style_override or style is None:
        if not style:
            style = {}
        result = {
            "questionmark": os.getenv("INQUIRERPY_STYLE_QUESTIONMARK", "#e5c07b"),
            "answermark": os.getenv("INQUIRERPY_STYLE_ANSWERMARK", "#e5c07b"),
            "answer": os.getenv("INQUIRERPY_STYLE_ANSWER", "#61afef"),
            "input": os.getenv("INQUIRERPY_STYLE_INPUT", "#98c379"),
            "question": os.getenv("INQUIRERPY_STYLE_QUESTION", ""),
            "answered_question": os.getenv("INQUIRERPY_STYLE_ANSWERED_QUESTION", ""),
            "instruction": os.getenv("INQUIRERPY_STYLE_INSTRUCTION", "#abb2bf"),
            "long_instruction": os.getenv(
                "INQUIRERPY_STYLE_LONG_INSTRUCTION", "#abb2bf"
            ),
            "pointer": os.getenv("INQUIRERPY_STYLE_POINTER", "#61afef"),
            "checkbox": os.getenv("INQUIRERPY_STYLE_CHECKBOX", "#98c379"),
            "separator": os.getenv("INQUIRERPY_STYLE_SEPARATOR", ""),
            "skipped": os.getenv("INQUIRERPY_STYLE_SKIPPED", "#5c6370"),
            "validator": os.getenv("INQUIRERPY_STYLE_VALIDATOR", ""),
            "marker": os.getenv("INQUIRERPY_STYLE_MARKER", "#e5c07b"),
            "fuzzy_prompt": os.getenv("INQUIRERPY_STYLE_FUZZY_PROMPT", "#c678dd"),
            "fuzzy_info": os.getenv("INQUIRERPY_STYLE_FUZZY_INFO", "#abb2bf"),
            "fuzzy_border": os.getenv("INQUIRERPY_STYLE_FUZZY_BORDER", "#4b5263"),
            "fuzzy_match": os.getenv("INQUIRERPY_STYLE_FUZZY_MATCH", "#c678dd"),
            "spinner_pattern": os.getenv("INQUIRERPY_STYLE_SPINNER_PATTERN", "#e5c07b"),
            "spinner_text": os.getenv("INQUIRERPY_STYLE_SPINNER_TEXT", ""),
            **style,
        }
    else:
        result = {
            "questionmark": os.getenv("INQUIRERPY_STYLE_QUESTIONMARK", ""),
            "answermark": os.getenv("INQUIRERPY_STYLE_ANSWERMARK", ""),
            "answer": os.getenv("INQUIRERPY_STYLE_ANSWER", ""),
            "input": os.getenv("INQUIRERPY_STYLE_INPUT", ""),
            "question": os.getenv("INQUIRERPY_STYLE_QUESTION", ""),
            "answered_question": os.getenv("INQUIRERPY_STYLE_ANSWERED_QUESTION", ""),
            "instruction": os.getenv("INQUIRERPY_STYLE_INSTRUCTION", ""),
            "long_instruction": os.getenv("INQUIRERPY_STYLE_LONG_INSTRUCTION", ""),
            "pointer": os.getenv("INQUIRERPY_STYLE_POINTER", ""),
            "checkbox": os.getenv("INQUIRERPY_STYLE_CHECKBOX", ""),
            "separator": os.getenv("INQUIRERPY_STYLE_SEPARATOR", ""),
            "skipped": os.getenv("INQUIRERPY_STYLE_SKIPPED", ""),
            "validator": os.getenv("INQUIRERPY_STYLE_VALIDATOR", ""),
            "marker": os.getenv("INQUIRERPY_STYLE_MARKER", ""),
            "fuzzy_prompt": os.getenv("INQUIRERPY_STYLE_FUZZY_PROMPT", ""),
            "fuzzy_info": os.getenv("INQUIRERPY_STYLE_FUZZY_INFO", ""),
            "fuzzy_border": os.getenv("INQUIRERPY_STYLE_FUZZY_BORDER", ""),
            "fuzzy_match": os.getenv("INQUIRERPY_STYLE_FUZZY_MATCH", ""),
            "spinner_pattern": os.getenv("INQUIRERPY_STYLE_SPINNER_PATTERN", ""),
            "spinner_text": os.getenv("INQUIRERPY_STYLE_SPINNER_TEXT", ""),
            **style,
        }

    if result.get("fuzzy_border"):
        result["frame.border"] = result.pop("fuzzy_border")
    if result.get("validator"):
        result["validation-toolbar"] = result.pop("validator")
    result["bottom-toolbar"] = "noreverse"
    return InquirerPyStyle(result)


def calculate_height(
    height: Optional[Union[int, str]],
    max_height: Optional[Union[int, str]],
    height_offset: int = 2,
) -> Tuple[Optional[int], int]:
    """Calculate the `height` and `max_height` for the main question contents.

    Tip:
        The parameter `height`/`max_height` can be specified by either a :class:`string` or :class:`int`.

        When `height`/`max_height` is :class:`str`:
            It will set the height to a percentage based on the value provided.
            You can optionally add the '%' sign which will be ignored while processing.

            Example: "60%" or "60" (60% of the current terminal visible lines)

        When `height`/`max_height` is :class:`int`:
            It will set the height to exact number of lines based on the value provided.

            Example: 20 (20 lines in terminal)

    Note:
        If `max_height` is not provided or is None, the default `max_height` will be configured to `70%` for
        best visual presentation in the terminal.

    Args:
        height: The desired height in either percentage as string or exact value as int.
        max_height: Maximum acceptable height in either percentage as string or exact value as int.
        height_offset: Height offset to apply to the height.

    Returns:
        A :class:`tuple` with the first value being the desired height and the second value being
        the maximum height.

    Raises:
        InvalidArgument: The provided `height`/`max_height` is not able to to be converted to int.

    Examples:
        >>> calculate_height(height="60%", max_height="100%")
    """
    try:
        _, term_lines = shutil.get_terminal_size()
        term_lines = term_lines
        if not height:
            dimmension_height = None
        else:
            if isinstance(height, str):
                height = height.replace("%", "")
                height = int(height)
                dimmension_height = (
                    math.floor(term_lines * (height / 100)) - height_offset
                )
            else:
                dimmension_height = height

        if not max_height:
            max_height = "70%" if not height else "100%"
        if isinstance(max_height, str):
            max_height = max_height.replace("%", "")
            max_height = int(max_height)
            dimmension_max_height = (
                math.floor(term_lines * (max_height / 100)) - height_offset
            )
        else:
            dimmension_max_height = max_height

        if dimmension_height and dimmension_height > dimmension_max_height:
            dimmension_height = dimmension_max_height
        if dimmension_height and dimmension_height <= 0:
            dimmension_height = 1
        if dimmension_max_height <= 0:
            dimmension_max_height = 1
        return dimmension_height, dimmension_max_height

    except ValueError:
        raise InvalidArgument(
            "prompt argument height/max_height needs to be type of an int or str"
        )


def patched_print(*values) -> None:
    """Patched :func:`print` that can print values without interrupting the prompt.

    See Also:
        :func:`print`
        :func:`~prompt_toolkit.application.run_in_terminal`

    Args:
        *values: Refer to :func:`print`.

    Examples:
        >>> patched_print("Hello World")
    """

    def _print():
        print(*values)

    run_in_terminal(_print)


def color_print(
    formatted_text: List[Tuple[str, str]], style: Optional[Dict[str, str]] = None
) -> None:
    """Print colored text leveraging :func:`~prompt_toolkit.shortcuts.print_formatted_text`.

    This function automatically handles printing the text without interrupting the
    current prompt.

    Args:
        formatted_text: A list of formatted_text.
        style: Style to apply to `formatted_text` in :class:`dictionary` form.

    Example:
        >>> color_print(formatted_text=[("class:aa", "hello "), ("class:bb", "world")], style={"aa": "red", "bb": "blue"})
        >>> color_print([("red", "yes"), ("", " "), ("blue", "no")])
    """

    def _print():
        print_formatted_text(
            FormattedText(formatted_text),
            style=Style.from_dict(style) if style else None,
        )

    if get_app().is_running:
        run_in_terminal(_print)
    else:
        _print()
